##############################################################################
#  
#  ZsapdbDA - a Database Adapter for Zope and sapdb
#  2001 by Ulrich Eck - net-labs Sytemhaus GmbH
#
##############################################################################

'''$Id: db.py,v 1.5 2002/11/21 10:28:53 jack Exp $'''
__version__='$Revision: 1.5 $'[11:-2]

import DCsapdb
from DateTime import DateTime
from Shared.DC.ZRDB.TM import TM

import string, sys
from string import strip, split, find
from time import time
from types import ListType

failures=0
calls=0
last_call_time=time()

class DB(TM):

    _p_oid=_p_changed=_registered=None

    Database_Connection=DCsapdb.connect
    Database_Error=DCsapdb.Error


    def __init__(self,connection):
        self.connection=connection
        self.kwargs = kwargs = self._parse_connection_string(connection)
        self.db=self.Database_Connection(kwargs.get('user',''),kwargs.get('password',''),
                                         kwargs.get('database',''),kwargs.get('host',''),
                                         timeout="0")
        self.cursor=self.db.cursor()
                
        
    def _parse_connection_string(self, connection):
        kwargs = {}
        items = split(connection)
        if not items: return kwargs
        db_host, items = items[0], items[1:]
        if '@' in db_host:
            db, host = split(db_host,'@',1)
            kwargs['database'] = db
            kwargs['host'] = host
        else:
            kwargs['database'] = db_host
        if not items: return kwargs
        kwargs['user'], items = items[0], items[1:]
        if not items: return kwargs
        kwargs['password'], items = items[0], items[1:]
        return kwargs


    def str(self,v, StringType=type('')):
        if v is None: return ''
        r=str(v)
        if r[-1:]=='L' and type(v) is not StringType: r=r[:-1]
        return r

    def _finish(self, *ignored):
        self.db.commit()

    def _abort(self, *ignored):
        self.db.rollback()

    def tables(self, rdb=0, all=0, _care=('TABLE', 'VIEW')):
        if all:
            t='ALL'
            e=', OWNER'
        else:
            t='USER'
            e=''
        c=self.cursor
        c.execute(
            'select OBJECT_NAME, OBJECT_TYPE%s from %s_OBJECTS' % (e,t)
            )
        r=[]
        a=r.append
        for name, typ in c.fetchall():
            if typ in _care:
                a({'TABLE_NAME': name, 'TABLE_TYPE': typ})
        return r

    def columns(self, table_name):
        c=self.cursor
        try: r=c.execute('select * from %s' % table_name)
        except: return ()
        desc=c.description
        r=[]
        a=r.append
        for name, typ, display_size, internal_size, precision, scale, null_ok in desc:
            a({ 'Name': name,
                'Type': typ,
                'Precision': precision,
                'Scale': scale,
                'Nullable': 0,
                })

        return r

    def query(self,query_string, max_rows=9999999):
        global failures, calls, last_call_time
        calls=calls+1
        desc=None
        result=[]
        self._register()
        c=self.cursor
        try:
            for qs in filter(None, map(strip,split(query_string, '\0'))):
                try:
                        r=c.execute(qs)
                except self.Database_Error, mess:
                        print mess.errorCode,mess.message
                        if mess.errorCode == 100:
                                r = None
                        else: raise self.Database_Error(mess)
                if r is not None:
                    if desc is not None:
                        if c.description != desc:
                            raise 'Query Error', (
                                'Multiple select schema are not allowed'
                                )
                        if max_rows:
                            for row in c.fetchmany(max_rows-len(result)):
                                result.append(row)
                    else:
                        desc=c.description
                        if max_rows and isinstance(r,DCsapdb.Cursor):
                            if max_rows==1: result=[c.fetchone()]
                            else: result=c.fetchmany(max_rows)
                                
            failures=0
            last_call_time=time()
        except self.Database_Error, mess:
            failures=failures+1
            if (failures > 1000 or time()-last_call_time > 600):
                # Hm. maybe the db is hosed.  Let's try once to restart it.
                failures=0
                last_call_time=time()
                c.close()
                self.db.close()
                db=self.db=self.Database_Connection(self.connection)
                self.cursor=db.cursor()
                return self.query(query_string, max_rows)
            else: raise sys.exc_type, sys.exc_value, sys.exc_traceback

        if desc is None: return (),()

        items=[]
        for i in range(len(desc)):
                name, typ, display_size, internal_size, precision, scale, null_ok = desc[i]
                if typ in ['Integer','Smallint']: typ='i'
                elif typ in ['Fixed','Float']: typ='n'
                elif typ in ['Date','Time','Timestamp']: 
                    for j in range(len(result)):
                            row = list(result[j])
                            if type( row[i] ) == type(""):
                                row[i] = DateTime(DCsapdb.valTranslation[typ]( row[i] ))
                                result[j] = tuple(row)
                    typ='d'
                elif typ in ['Long','Long Varchar']: 
                    for j in range(len(result)):
                        row = list(result[j])
                        if row[i] is not None:
                            row[i] = _longConvertFromDB( row[i] )
                            result[j] = tuple(row)
                    typ='s'
                else: typ='s'
                items.append({
                        'name': name,
                        'type': typ,
                        'width': display_size,
                        'null': 0,
                        })
        return items, tuple(result)

def _longConvertFromDB(val):
    ret = ''
    while 1:
        d = val.read(1000)
        if len(d) == 0:
            break
        ret = ret + d
    return ret

